//----------------------------------------------------------------------------------//
//                                                                                  //
//  Created on November 09, 2006                                                    //
//                                                                                  // 
//  "Copyright (c) 2006 The Regents of the University of California.                //
//  All rights reserved.                                                            //
//                                                                                  //
//  Permission to use, copy, modify, and distribute this software and its           //
//  documentation for any purpose, without fee, and without written agreement is    //
//  hereby granted, provided that the above copyright notice, the following         //
//  two paragraphs and the author appear in all copies of this software.            //
//                                                                                  //
//  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR       //
//  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT     //
//  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF    //
//  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                  //
//                                                                                  //
//  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,             //
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY        //
//  AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS       //
//  ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO      //
//  PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."         //
//                                                                                  //
//----------------------------------------------------------------------------------//


//----------------------------------------------------------------------------------
//
// chardevice.cpp
// 
//    Cross-platform one-to-one communication over serial, TCP (client and server),
//    and UDP (client and server) through a simple read/write interface that is
//    robust to disconnection. Also a "wrapper" char device that allows any one of
//    the above to be used transparently, and a "buffered character device" that
//    serves as a base class for message-passing architectures that receive a single
//    type of message that is no more than 256 characters in length (transparently
//    over any one of the supported interfaces).
//
// authors
//
//    Todd R. Templeton <ttemplet@eecs.berkeley.edu>
//    David H. Shim <hcshim@eecs.berkeley.edu> (QNX serial initialization)
//
//----------------------------------------------------------------------------------


#ifndef __CHARDEVICE_C__
#define __CHARDEVICE_C__

#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
#include <Windows.h>
#include <winsock.h>
#include <sys/timeb.h>
#include <time.h>
#else
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>  
#include <unistd.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/time.h>
#ifndef SOL_TCP
/* for ifndef SOL_TCP below */
#include <netdb.h>
#endif
#endif

#include "chardevice.h"


//#define CHARDEVICE_MAIN


#ifdef WIN32
typedef SOCKADDR sockaddr;
typedef SOCKADDR_IN sockaddr_in;
typedef int socklen_t;
#else
#define SOCKET_ERROR -1
#ifndef SOL_TCP
#define SOL_TCP getprotobyname("TCP")->p_proto
#endif
#ifndef MSG_NOSIGNAL
/* FreeBSD and OS X use SO_NOSIGPIPE instead */
#define MSG_NOSIGNAL SO_NOSIGPIPE
#endif
#ifndef ENONET
/* always use these two together */
#define ENONET EHOSTDOWN
#endif
/* ensure that either both or neither of {VSWTC,VSWTCH} are set */
#if defined(VSWTCH) && !defined(VSWTC)
#define VSWTC VSWTCH
#endif
#if defined(VSWTC) && !defined(VSWTCH)
#define VSWTCH VSWTC
#endif
#endif


#ifdef WIN32

// get elapsed time, first call returns 0.0
double CharDevice::get_time( void )
{
    static bool getTimeInit = false;
    static struct _timeb tvStart;
    struct _timeb tv;

    if( !getTimeInit )
    {
        _ftime( &tvStart );
        getTimeInit = true;
    }
        
    _ftime( &tv );
    return ( (double)tv.time - (double)tvStart.time ) + 0.001 * ( (double)tv.millitm - (double)tvStart.millitm );
}

#else

// get elapsed time, first call returns 0.0
double CharDevice::get_time( void )
{
    static bool getTimeInit = false;
    static struct timeval tvStart;
    struct timeval tv;
    struct timezone tz;

    if( !getTimeInit )
    {
        gettimeofday( &tvStart, &tz );
        getTimeInit = true;
    }
        
    gettimeofday( &tv, &tz );
    return ( (double)tv.tv_sec - (double)tvStart.tv_sec ) + 0.000001 * ( (double)tv.tv_usec - (double)tvStart.tv_usec );
}

#endif


#ifdef WIN32

// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
// for info about win32 serial API, see http://www.robbayer.com/files/serial-win.pdf
int SerialCharDevice::open(bool wait /* = true */)
{
    int i;
    int len = strlen( options );
    char *szport;
    int baudrate;
    DCB params = {0};
    COMMTIMEOUTS timeouts = {0};
    int baudmask;
    
    // find serial port device
    for( i = 0; i < len && options[ i ] != ','; i++ );
    assert( ( i + 1 ) < len );
    options[ i ] = '\0';
    szport = options;
    // find baud rate
    baudrate = atoi( &options[ i + 1 ] );

    fprintf(stderr, "InitSer: opening %s with baudrate %d...\n", szport, baudrate);
    fd = CreateFile(szport, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0 ); 
    if (fd == INVALID_HANDLE_VALUE) {
        char lastError[1024];
        FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), lastError, 1024, NULL);
        fputs(lastError, stderr);
        exit(-1);
    }
    
    switch(baudrate)
    {
    case 110:
        baudmask = CBR_110;
        break;
    case 300:
        baudmask = CBR_300;
        break;
    case 600:
        baudmask = CBR_600;
        break;
    case 1200:
        baudmask = CBR_1200;
        break;
    case 2400:
        baudmask = CBR_2400;
        break;
    case 4800:
        baudmask = CBR_4800;
        break;
    case 9600:
        baudmask = CBR_9600;
        break;
    case 19200:
        baudmask = CBR_19200;
        break;
    case 38400:
        baudmask = CBR_38400;
        break;
    case 57600:
        baudmask = CBR_57600;
        break;
    case 115200:
        baudmask = CBR_115200;
        break;
    default:
        fprintf(stderr, "Unknown baudrate %d\r\n", baudrate);
        baudmask = CBR_38400;
    }

    params.DCBlength = sizeof(params);
    if(!GetCommState(fd, &params)) {
        char lastError[1024];
        FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), lastError, 1024, NULL);
        fputs(lastError, stderr);
        exit(-1);
    }
    params.BaudRate = baudmask;
    params.ByteSize = 8;
    params.StopBits = ONESTOPBIT;
    params.Parity   = NOPARITY;
    if(!SetCommState(fd, &params)) {
        char lastError[1024];
        FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), lastError, 1024, NULL);
        fputs(lastError, stderr);
        exit(-1);
    }
    
    timeouts.ReadIntervalTimeout     = MAXDWORD;
    timeouts.ReadTotalTimeoutConstant    = 0;
    timeouts.ReadTotalTimeoutMultiplier  = 0;
    timeouts.WriteTotalTimeoutConstant   = 50;
    timeouts.WriteTotalTimeoutMultiplier = 10;
    if(!SetCommTimeouts(fd, &timeouts)) {
        char lastError[1024];
        FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), lastError, 1024, NULL);
        fputs(lastError, stderr);
        exit(-1);
    }
    
    // repair options
    options[ i ] = ',';
    
    return fd == INVALID_HANDLE_VALUE ? -1 : 0;
}

#elif defined(__QNX__)

// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
// Serial initialization routine from David Shim
int SerialCharDevice::open(bool wait /* = true */)
{
    int i;
    int len = strlen( options );
    char *szport;
    int baudrate;
    //int fd;
    struct termios termios_p;
    int baudmask;
    char buf[30];
    
    // find serial port device
    for( i = 0; i < len && options[ i ] != ','; i++ );
    assert( ( i + 1 ) < len );
    options[ i ] = '\0';
    szport = options;
    // find baud rate
    baudrate = atoi( &options[ i + 1 ] );

    fd = open( szport,  O_RDWR); /* open ser*/
    if ( fd <= NULL ) {
        fprintf( stderr, "Error opening device %s for input/output\r\n", szport );
        exit( -1 );
    }
    tcgetattr(fd, &termios_p);         /*load termios_p w/ current params */

    switch(baudrate)
    {
    case 0:
        baudmask = B0;
        break;
    case 50:
        baudmask = B50;
        break;
    case 75:
        baudmask = B75;
        break;
    case 110:
        baudmask = B110;
        break;
    case 134:
        baudmask = B134;
        break;
    case 150:
        baudmask = B150;
        break;
    case 200:
        baudmask = B200;
        break;
    case 300:
        baudmask = B300;
        break;
    case 600:
        baudmask = B600;
        break;
    case 1200:
        baudmask = B1200;
        break;
    case 1800:
        baudmask = B1800;
        break;
    case 2400:
        baudmask = B2400;
        break;
    case 4800:
        baudmask = B4800;
        break;
    case 9600:
        baudmask = B9600;
        break;
    case 19200:
        baudmask = B19200;
        break;
    case 38400:
        baudmask = B38400;
        break;
    case 57600:
        baudmask = B57600;
        break;
    case 115200:
        baudmask = B115200;
        break;
    case 230400:
        baudmask = B230400;
        break;
    default:
        fprintf(stderr, "Unknown baudrate %d\r\n", baudrate);
    }
    
    cfsetispeed(&termios_p, baudmask);     /* set baud rates*/
    cfsetospeed(&termios_p, baudmask);
    termios_p.c_cflag = CS8|CREAD;//|PARENB|PARODD; // no flow control |IHFLOW|OHFLOW;
    tcsetattr(fd, TCSANOW, &termios_p);     
    dev_mode(fd, 0, _DEV_MODES);
    dev_state(fd, 0, _DEV_EVENT_INPUT);
    tcflush(fd,TCIOFLUSH);
    while (dev_ischars(fd))
        dev_read(fd,buf,sizeof(buf),sizeof(buf),0,0,0,0);
        
    // repair options
    options[ i ] = ',';

    return(fd);

}

#else

// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
int SerialCharDevice::open(bool wait /* = true */)
{
    int i;
    int len = strlen( options );
    char *szport;
    int baudrate;
    //int fd;
    struct termios oldtio,newtio;
    int baudmask;
    
    // find serial port device
    for( i = 0; i < len && options[ i ] != ','; i++ );
    assert( ( i + 1 ) < len );
    options[ i ] = '\0';
    szport = options;
    // find baud rate
    baudrate = atoi( &options[ i + 1 ] );

    fprintf(stderr, "InitSer: opening %s with baudrate %d...\n", szport, baudrate);
    fd = ::open(szport, O_RDWR | O_NOCTTY ); 
    if (fd <0) {perror(szport); exit(-1); }

    tcgetattr(fd,&oldtio); /* save current serial port settings */
    bzero(&newtio, sizeof(newtio)); /* clear struct for new port settings */

    newtio.c_cflag = /*baudrate  |*/ CS8 | CLOCAL | CREAD /*| CRTSCTS*/;
    newtio.c_iflag = IGNPAR; // | ICRNL; ->IGNORE CARRIGE RETURN, WHICH IS A BIG NO-NO FOR BINARY STREAM
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;//ICANON; // DISABLED ANYTHING SUSPICIOUS AND SEEM UNNECESSARY
    newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */ 
    newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
    newtio.c_cc[VERASE]   = 0;     /* del */
    newtio.c_cc[VKILL]    = 0;     /* @ */
    newtio.c_cc[VEOF]     = 0;     /* Ctrl-d */
    newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
    newtio.c_cc[VMIN]     = 0;     /* blocking read until 0 characters arrive (non-blocking) */
    #ifdef VSWTC
    /* ensured above that either both or neither are defined */
    newtio.c_cc[VSWTC]    = 0;     /* '\0' */
    newtio.c_cc[VSWTCH]   = 0;     /* '\0' */
    #endif
    newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */ 
    newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
    newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
    newtio.c_cc[VEOL]     = 0;     /* '\0' */
    newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
    newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
    newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
    newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
    newtio.c_cc[VEOL2]    = 0;     /* '\0' */
    
    switch(baudrate)
    {
    case 0:
        baudmask = B0;
        break;
    case 50:
        baudmask = B50;
        break;
    case 75:
        baudmask = B75;
        break;
    case 110:
        baudmask = B110;
        break;
    case 134:
        baudmask = B134;
        break;
    case 150:
        baudmask = B150;
        break;
    case 200:
        baudmask = B200;
        break;
    case 300:
        baudmask = B300;
        break;
    case 600:
        baudmask = B600;
        break;
    case 1200:
        baudmask = B1200;
        break;
    case 1800:
        baudmask = B1800;
        break;
    case 2400:
        baudmask = B2400;
        break;
    case 4800:
        baudmask = B4800;
        break;
    case 9600:
        baudmask = B9600;
        break;
    case 19200:
        baudmask = B19200;
        break;
    case 38400:
        baudmask = B38400;
        break;
    case 57600:
        baudmask = B57600;
        break;
    case 115200:
        baudmask = B115200;
        break;
    case 230400:
        baudmask = B230400;
        break;
    default:
        fprintf(stderr, "Unknown baudrate %d\r\n", baudrate);
        baudmask = B38400;
    }
    
    cfsetispeed(&newtio, baudmask);
    cfsetospeed(&newtio, baudmask);
    
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);
    
    // repair options
    options[ i ] = ',';
    
    return fd;
}

#endif

// read from device (buf = buffer into which to read, count = max number of bytes to read,
//   maxInterval = time in seconds since last successful read after which to determine that
//   link has gone down, wait = whether to wait for reconnection if link has gone down,
//   returns number of bytes read or -1 on error)
ssize_t SerialCharDevice::read( void *buf, size_t count, double maxInterval /* = -1.0 */, bool wait /* = true */ )
{
    //fprintf( stderr, "Reading up to %d bytes\n", count );
    #ifdef WIN32
    DWORD retval;
    if( !ReadFile( fd, buf, count, &retval, NULL ) )
        return -1;
    return retval;
    #else
    return ::read( fd, buf, count );
    #endif
}

// write to device (buf = buffer from which to write, count = number number of bytes to write,
//   wait = whether to wait for reconnection if link has gone down, returns number of bytes
//   written or -1 on error)
ssize_t SerialCharDevice::write( const void *buf, size_t count, bool wait /*= true*/ )
{
    //fprintf( stderr, "Writing %d bytes\n", count );
    #ifdef WIN32
    DWORD retval;
    if( !WriteFile( fd, buf, count, &retval, NULL ) )
        return -1;
    return retval;
    #else
    return ::write( fd, buf, count );
    #endif
}

#ifdef WIN32

// close device, called last time by destructor (returns 0 on success or -1 on error)
int SerialCharDevice::close( void )
{
    int retval = 0;
    if( fd != INVALID_HANDLE_VALUE )
    {
        retval = CloseHandle( fd );
        retval = ( retval == 0 ) ? -1 : 0;
        fd = INVALID_HANDLE_VALUE;
    }
    return retval;
}

#else

// close device, called last time by destructor (returns 0 on success or -1 on error)
int SerialCharDevice::close( void )
{
    int retval = 0;
    if( fd != INVALID_HANDLE_VALUE )
    {
        retval = ::close( fd );
        fd = INVALID_HANDLE_VALUE;
    }
    return retval;
}

#endif


// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
int TcpCharDevice::open( bool wait /* = true */ )
{
    int i/*, j*/;
    int len = strlen( options );
    char *host;
    int port;
    sockaddr_in addr;
    int nodelay = 1;
    #ifdef WIN32
    unsigned long nonblocking = 1;
    int lerror = 0;
    #endif
    bool success = false;
    int retval;
    
    // find host
    for( i = 0; i < len && options[ i ] != ','; i++ );
    assert( ( i + 1 ) < len );
    options[ i ] = '\0';
    host = options;
    // find port
    port = atoi( &options[ i + 1 ] );
    
    while( 1 )
    {
        if( !connecting )
        {
            fprintf( stderr, "connecting to %s port %d\n", host, port );
        
            #ifdef WIN32
            if( !this->startedWSA )
            {
                WORD versionRequested = MAKEWORD( 1, 1 );
                WSADATA wd;
                WSAStartup( versionRequested, &wd );
                assert( wd.wVersion == versionRequested );
                this->startedWSA = true;
            }
            #endif
    
            connectsock = socket( PF_INET, SOCK_STREAM, 0 );
            assert( connectsock != INVALID_SOCKET );
            addr.sin_family = AF_INET;
            addr.sin_port = htons( port );
            #ifdef WIN32
            addr.sin_addr.s_addr = inet_addr( host );
            #else
            retval = inet_pton( AF_INET, host, &addr.sin_addr );
            assert( retval > 0 );
            #endif
            // set nonblocking
            #ifdef WIN32
            retval = ioctlsocket( connectsock, FIONBIO, &nonblocking );
            #else
            retval = fcntl( connectsock, F_SETFL, fcntl( connectsock, F_GETFL ) | O_NONBLOCK );
            #endif
            assert( retval != SOCKET_ERROR );
            retval = connect( connectsock, (sockaddr*)&addr, sizeof( addr ) );
            #ifdef WIN32
            if( retval == SOCKET_ERROR )
                lerror = WSAGetLastError();
            if( retval != SOCKET_ERROR || lerror == WSAEWOULDBLOCK )
                connecting = true;
            #else
            if( retval != SOCKET_ERROR || errno == EINPROGRESS /*|| errno == 0*/ ) //FIXME: on stargate, errno is 0
                connecting = true;
            #endif
            else
            {
                connecting = false;
                this->close_connection( );
            }
        }
        if( connecting )
        {
            while( 1 )
            {
                struct timeval tv;
                fd_set wset, eset;
                #ifndef WIN32
                int sockoptVal;
                socklen_t sockoptValSize = sizeof( int );
                #endif
                FD_ZERO( &wset );
                FD_SET( connectsock, &wset );
                FD_ZERO( &eset );
                #ifdef WIN32
                FD_SET( connectsock, &eset );
                #endif
                tv.tv_sec = 0;
                tv.tv_usec = 0;
                if( select( connectsock + 1, 0, &wset, &eset, &tv ) <= 0 || ( !FD_ISSET( connectsock, &wset ) && !FD_ISSET( connectsock, &eset ) ) )
                {
                    if( wait )
                    {
                        #ifdef WIN32
                        Sleep( 100 );
                        #else
                        usleep( 100000 );
                        #endif
                        continue;
                    }
                    else
                        break;
                }
                #ifdef WIN32
                else if( FD_ISSET( connectsock, &eset ) )
                #else
                else if( getsockopt( connectsock, SOL_SOCKET, SO_ERROR, &sockoptVal, &sockoptValSize ) < 0 || sockoptVal != 0 )
                #endif
                {
                    this->close( );
                    connecting = false;
                    break;
                }
                fd = connectsock;
                connectsock = INVALID_SOCKET;
                connecting = false;
                // set nodelay
                #ifdef WIN32
                retval = setsockopt( fd, IPPROTO_TCP, TCP_NODELAY, (const char*)&nodelay, sizeof( nodelay ) );
                #else
                retval = setsockopt( fd, SOL_TCP, TCP_NODELAY, (const void*)&nodelay, sizeof( nodelay ) );
                #endif
                assert( retval != SOCKET_ERROR );
                this->lastReadTime = this->get_time( );
                fprintf( stderr, "...connected\n" );
                success = true;
                break;
            }
        }
        if( success )
            break;
        else if( wait )
        {
            #ifdef WIN32
            Sleep( 100 );
            #else
            usleep( 100000 );
            #endif
            continue;
        }
        else
            break;
    }
    /*if( retval == SOCKET_ERROR && wait )
    {
        fprintf( stderr, "Unable to connect to host %s port %d\n", host, port );
        exit( EXIT_FAILURE );
    }*/
    //if( retval == SOCKET_ERROR )
    //    this->close( );
    
    // repair options
    options[ i ] = ',';
    
    return fd;
}

// read from device (buf = buffer into which to read, count = max number of bytes to read,
//   maxInterval = time in seconds since last successful read after which to determine that
//   link has gone down, wait = whether to wait for reconnection if link has gone down,
//   returns number of bytes read or -1 on error)
ssize_t TcpCharDevice::read( void *buf, size_t count, double maxInterval /* = -1.0 */, bool wait /* = true */ )
{
    ssize_t retval;
    while( 1 )
    {
        #ifdef WIN32
        int lerror;
        retval = ::recv( fd, (char*)buf, count, 0 ); //NOTE: was ::read()
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( ( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #else
        retval = ::recv( fd, buf, count, 0 ); //NOTE: was ::read()
        if( ( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #endif
    }
    if( retval == SOCKET_ERROR )
        retval = -1;
    if( retval > 0 )
        this->lastReadTime = this->get_time( );
    return retval;
}

// write to device (buf = buffer from which to write, count = number number of bytes to write,
//   wait = whether to wait for reconnection if link has gone down, returns number of bytes
//   written or -1 on error)
ssize_t TcpCharDevice::write( const void *buf, size_t count, bool wait /*= true*/ )
{
    ssize_t retval;
    while( 1 )
    {
        #ifdef WIN32
        int lerror;
        retval = ::send( fd, (char*)buf, count, 0 );
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #else
        retval = ::send( fd, buf, count, MSG_NOSIGNAL );
        if( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #endif
    }
    #ifdef WIN32
    if( retval == SOCKET_ERROR )
        retval = -1;
    #endif
    return retval;
}

// close device, called last time by destructor (returns 0 on success or -1 on error)
int TcpCharDevice::close( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    if( connectsock != INVALID_SOCKET )
    {
        #ifdef WIN32
        closesocket( connectsock );
        #else
        ::close( connectsock );
        #endif
        connectsock = INVALID_SOCKET;
    }
    connecting = false;
    #ifdef WIN32
    if( this->startedWSA )
    {
        WSACleanup( );
        this->startedWSA = false;
    }
    #endif
    return retval;
}

// close connection (returns 0 on success or -1 on error)
int TcpCharDevice::close_connection( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    if( connectsock != INVALID_SOCKET && !connecting )
    {
        #ifdef WIN32
        closesocket( connectsock );
        #else
        ::close( connectsock );
        #endif
        connectsock = INVALID_SOCKET;
    }
    return retval;
}


// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
int TcpServerCharDevice::open( bool wait /* = true */ )
{
    int i;
    int len = strlen( options );
    char *client;
    char defaultmask[16] = "255.255.255.255";
    char *mask = defaultmask;
    int maskpos = -1;
    int port;
    //int listensock;
    sockaddr_in addr, addr2, addrmask;
    socklen_t addrlen;
    int reuseaddr = 1;
    int nodelay = 1;
    #ifdef WIN32
    unsigned long nonblocking = 1;
    #endif
    int retval;
    
    // find client
    for( i = 0; i < len && options[ i ] != ','; i++ )
    {
        if( options[ i ] == '/' && ( i + 1 ) < len && maskpos < 0 )
        {
            options[ i ] = '\0';
            maskpos = i + 1;
            mask = &options[ maskpos ];
        }
    }
    assert( ( i + 1 ) < len );
    options[ i ] = '\0';
    client = options;
    // find port
    port = atoi( &options[ i + 1 ] );
    
    addrlen = sizeof( addr );
    if( !listening )
    {
        fprintf( stderr, "opening port %d to listen for connections, waiting for client %s/%s\n", port, client, mask );
        
        #ifdef WIN32
        if( !this->startedWSA )
        {
            WORD versionRequested = MAKEWORD( 1, 1 );
            WSADATA wd;
            WSAStartup( versionRequested, &wd );
            assert( wd.wVersion == versionRequested );
            this->startedWSA = true;
        }
        #endif
        
        listensock = socket( PF_INET, SOCK_STREAM, 0 );
        assert( listensock != INVALID_SOCKET );
        // set nonblocking
        #ifdef WIN32
        retval = ioctlsocket( listensock, FIONBIO, &nonblocking );
        #else
        retval = fcntl( listensock, F_SETFL, fcntl( listensock, F_GETFL ) | O_NONBLOCK );
        #endif
        assert( retval != SOCKET_ERROR );
        #ifdef WIN32
        retval = setsockopt( listensock, SOL_SOCKET, SO_REUSEADDR, (const char*)&reuseaddr, sizeof( reuseaddr ) );
        #else
        retval = setsockopt( listensock, SOL_SOCKET, SO_REUSEADDR, (const void*)&reuseaddr, sizeof( reuseaddr ) );
        #endif
        assert( retval != SOCKET_ERROR );
        addr.sin_family = AF_INET;
        addr.sin_port = htons( port );
        addr.sin_addr.s_addr = INADDR_ANY;
        retval = bind( listensock, (sockaddr*)&addr, addrlen );
        assert( retval != SOCKET_ERROR );
        retval = listen( listensock, 1 );
        assert( retval != SOCKET_ERROR );
        /*#if 0
        fd = accept( listensock, (sockaddr*)&addr, &addrlen );
        assert( fd != -1 );
        retval = inet_pton( AF_INET, client, &addr2.sin_addr );
        assert( retval > 0 );
        assert( addr.sin_addr.s_addr == addr2.sin_addr.s_addr ); // check address of client
        #else
        retval = inet_pton( AF_INET, client, &addr2.sin_addr );
        assert( retval > 0 );
        #endif
        */
        listening = true;
    }
    #ifdef WIN32
    addr2.sin_addr.s_addr = inet_addr( client );
    addrmask.sin_addr.s_addr = inet_addr( mask );
    #else
    retval = inet_pton( AF_INET, client, &addr2.sin_addr );
    assert( retval > 0 );
    retval = inet_pton( AF_INET, mask, &addrmask.sin_addr );
    assert( retval > 0 );
    #endif
    while( 1 )
    {
        fd = accept( listensock, (sockaddr*)&addr, &addrlen );
        #ifdef WIN32
        int lerror = WSAGetLastError();
        assert( fd != INVALID_SOCKET || lerror == WSAECONNRESET || lerror == WSAENETDOWN || lerror == WSAEWOULDBLOCK );
        if( fd == INVALID_SOCKET && ( lerror == WSAECONNRESET || lerror == WSAENETDOWN || lerror == WSAEWOULDBLOCK ) )
        {
            if( wait )
            {
                Sleep( 100 );
                continue;
            }
            else
                break;
        }
        #else
        assert( fd != INVALID_SOCKET || errno == EAGAIN || errno == ENETDOWN || errno == EPROTO || errno == ENOPROTOOPT || errno == EHOSTDOWN || errno == ENONET || errno == EHOSTUNREACH || errno == EOPNOTSUPP || errno == ENETUNREACH );
        if( fd == INVALID_SOCKET && ( errno == EAGAIN || errno == ENETDOWN || errno == EPROTO || errno == ENOPROTOOPT || errno == EHOSTDOWN || errno == ENONET || errno == EHOSTUNREACH || errno == EOPNOTSUPP || errno == ENETUNREACH ) )
        {
            if( wait )
            {
                usleep( 100000 );
                continue;
            }
            else
                break;
        }
        #endif
        if( !( ( addr.sin_addr.s_addr ^ addr2.sin_addr.s_addr ) & addrmask.sin_addr.s_addr ) ) // check address of client
        {
            // set nonblocking
            #ifdef WIN32
            retval = ioctlsocket( fd, FIONBIO, &nonblocking );
            #else
            retval = fcntl( fd, F_SETFL, fcntl( fd, F_GETFL ) | O_NONBLOCK );
            #endif
            assert( retval != SOCKET_ERROR );
            // set nodelay
            #ifdef WIN32
            retval = setsockopt( fd, IPPROTO_TCP, TCP_NODELAY, (const char*)&nodelay, sizeof( nodelay ) );
            #else
            retval = setsockopt( fd, SOL_TCP, TCP_NODELAY, (const void*)&nodelay, sizeof( nodelay ) );
            #endif
            assert( retval != SOCKET_ERROR );
            //::close( listensock );
            this->lastReadTime = this->get_time( );
            fprintf( stderr, "...connected\n" );
            break;
        }
        else {
            fprintf( stderr, "rejecting connection from incorrect client\n" );
            this->close_client( );
        }
    }
    
    // repair options
    if( maskpos > 0 )
        options[ maskpos - 1 ] = '/';
    options[ i ] = ',';
    
    return fd;
}

// read from device (buf = buffer into which to read, count = max number of bytes to read,
//   maxInterval = time in seconds since last successful read after which to determine that
//   link has gone down, wait = whether to wait for reconnection if link has gone down,
//   returns number of bytes read or -1 on error)
ssize_t TcpServerCharDevice::read( void *buf, size_t count, double maxInterval /* = -1.0 */, bool wait /* = true */ )
{
    ssize_t retval;
    while( 1 )
    {
        #ifdef WIN32
        int lerror;
        retval = ::recv( fd, (char*)buf, count, 0 ); //NOTE: was ::read()
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( ( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #else
        retval = ::recv( fd, buf, count, 0 ); //NOTE: was ::read()
        if( ( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #endif
    }
    if( retval == SOCKET_ERROR )
        retval = -1;
    if( retval > 0 )
        this->lastReadTime = this->get_time( );
    return retval;
}

// write to device (buf = buffer from which to write, count = number number of bytes to write,
//   wait = whether to wait for reconnection if link has gone down, returns number of bytes
//   written or -1 on error)
ssize_t TcpServerCharDevice::write( const void *buf, size_t count, bool wait /*= true*/ )
{
    ssize_t retval;
    while( 1 )
    {
        #ifdef WIN32
        int lerror;
        retval = ::send( fd, (char*)buf, count, 0 );
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #else
        retval = ::send( fd, buf, count, MSG_NOSIGNAL );
        if( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #endif
    }
    #ifdef WIN32
    if( retval == SOCKET_ERROR )
        retval = -1;
    #endif
    return retval;
}

// close device, called last time by destructor (returns 0 on success or -1 on error)
int TcpServerCharDevice::close( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    if( listensock != INVALID_SOCKET )
    {
        #ifdef WIN32
        closesocket( listensock );
        #else
        ::close( listensock );
        #endif
        listensock = INVALID_SOCKET;
    }
    listening = false;
    #ifdef WIN32
    if( this->startedWSA )
    {
        WSACleanup( );
        this->startedWSA = false;
    }
    #endif
    return retval;
}

// close current client (returns 0 on success or -1 on error)
int TcpServerCharDevice::close_client( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    return retval;
}

// whether currently connected
bool TcpServerCharDevice::connected( void )
{
    return ( fd != INVALID_SOCKET );
}

// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
int UdpCharDevice::open( bool wait /* = true */ )
{
    int i/*, j*/;
    int len = strlen( options );
    char *host;
    int port;
    sockaddr_in addr;
    //int nodelay = 1;
    #ifdef WIN32
    unsigned long nonblocking = 1;
    #endif
    int retval;
    
    // find host
    for( i = 0; i < len && options[ i ] != ','; i++ );
    assert( ( i + 1 ) < len );
    options[ i ] = '\0';
    host = options;
    // find port
    port = atoi( &options[ i + 1 ] );
    
    if( fd == INVALID_SOCKET )
    {
        fprintf( stderr, "connecting to %s port %d\n", host, port );
    
        #ifdef WIN32
        if( !this->startedWSA )
        {
            WORD versionRequested = MAKEWORD( 1, 1 );
            WSADATA wd;
            WSAStartup( versionRequested, &wd );
            assert( wd.wVersion == versionRequested );
            this->startedWSA = true;
        }
        #endif

        fd = socket( PF_INET, SOCK_DGRAM, 0 );
        assert( fd != INVALID_SOCKET );
        addr.sin_family = AF_INET;
        addr.sin_port = htons( port );
        #ifdef WIN32
        addr.sin_addr.s_addr = inet_addr( host );
        #else
        retval = inet_pton( AF_INET, host, &addr.sin_addr );
        assert( retval > 0 );
        #endif
        // set nonblocking
        #ifdef WIN32
        retval = ioctlsocket( fd, FIONBIO, &nonblocking );
        #else
        retval = fcntl( fd, F_SETFL, fcntl( fd, F_GETFL ) | O_NONBLOCK );
        #endif
        assert( retval != SOCKET_ERROR );
        retval = connect( fd, (sockaddr*)&addr, sizeof( addr ) );
        assert( retval != SOCKET_ERROR );
        this->lastReadTime = this->get_time( );
        fprintf( stderr, "...connected\n" );
    }
    
    // repair options
    options[ i ] = ',';
    
    return fd;
}

// read from device (buf = buffer into which to read, count = max number of bytes to read,
//   maxInterval = time in seconds since last successful read after which to determine that
//   link has gone down, wait = whether to wait for reconnection if link has gone down,
//   returns number of bytes read or -1 on error)
//NOTE: SOCK_DGRAM does not return everything in buffer, only last message; this function compensates for that by looping the recv (but does not read part of a message, which would cause the rest of the message to be discarded)
ssize_t UdpCharDevice::read( void *buf, size_t count, double maxInterval /* = -1.0 */, bool wait /* = true */ )
{
    ssize_t retval = 0;
    ssize_t sumretval = 0;
    bool peek = false;
    while( (size_t)sumretval < count )
    {
        #ifdef WIN32
        int lerror;
        retval = ::recv( fd, &(((char*)buf)[sumretval]), count - sumretval, peek ? MSG_PEEK : 0 ); //NOTE: was ::read()
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( ( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
            continue;
        }
        #else
        retval = ::recv( fd, (void*)&(((char*)buf)[sumretval]), count - sumretval, peek ? MSG_PEEK : 0 ); //NOTE: was ::read()
        if( ( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
            continue;
        }
        #endif
        if( retval <= 0 )
            break;
        if( peek )
        {
            if( (size_t)retval > ( count - (size_t)sumretval ) )
                break;
        }
        else
        { // data is OK
            sumretval += retval;
        }
        peek = !peek;
    }
    if( sumretval > 0 )
        retval = sumretval;
    else if( retval == SOCKET_ERROR )
        retval = -1;
    if( retval > 0 )
        this->lastReadTime = this->get_time( );
    return retval;
}

// write to device (buf = buffer from which to write, count = number number of bytes to write,
//   wait = whether to wait for reconnection if link has gone down, returns number of bytes
//   written or -1 on error)
ssize_t UdpCharDevice::write( const void *buf, size_t count, bool wait /*= true*/ )
{
    ssize_t retval;
    while( 1 )
    {
        #ifdef WIN32
        int lerror;
        retval = ::send( fd, (char*)buf, count, 0 );
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #else
        retval = ::send( fd, buf, count, MSG_NOSIGNAL );
        if( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) )
        {
            this->close_connection( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #endif
    }
    #ifdef WIN32
    if( retval == SOCKET_ERROR )
        retval = -1;
    #endif
    return retval;
}

// close device, called last time by destructor (returns 0 on success or -1 on error)
int UdpCharDevice::close( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    #ifdef WIN32
    if( this->startedWSA )
    {
        WSACleanup( );
        this->startedWSA = false;
    }
    #endif
    return retval;
}

// close connection (returns 0 on success or -1 on error)
int UdpCharDevice::close_connection( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    return retval;
}


// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
int UdpServerCharDevice::open( bool wait /* = true */ )
{
    int i;
    int len = strlen( options );
    char *client;
    char defaultmask[16] = "255.255.255.255";
    char *mask = defaultmask;
    int maskpos = -1;
    int port;
    //int listensock;
    sockaddr_in addr, addr2, addrmask;
    socklen_t addrlen;
    int reuseaddr = 1;
    //int nodelay = 1;
    #ifdef WIN32
    unsigned long nonblocking = 1;
    #endif
    int retval;
    
    // find client
    for( i = 0; i < len && options[ i ] != ','; i++ )
    {
        if( options[ i ] == '/' && ( i + 1 ) < len && maskpos < 0 )
        {
            options[ i ] = '\0';
            maskpos = i + 1;
            mask = &options[ maskpos ];
        }
    }
    assert( ( i + 1 ) < len );
    options[ i ] = '\0';
    client = options;
    // find port
    port = atoi( &options[ i + 1 ] );
    
    addrlen = sizeof( addr );

    if( fd == INVALID_SOCKET )
    {
        fprintf( stderr, "opening port %d to listen for connections, waiting for client %s/%s\n", port, client, mask );
        
        #ifdef WIN32
        if( !this->startedWSA )
        {
            WORD versionRequested = MAKEWORD( 1, 1 );
            WSADATA wd;
            WSAStartup( versionRequested, &wd );
            assert( wd.wVersion == versionRequested );
            this->startedWSA = true;
        }
        #endif
        
        fd = socket( PF_INET, SOCK_DGRAM, 0 );
        assert( fd != INVALID_SOCKET );
        // set nonblocking
        #ifdef WIN32
        retval = ioctlsocket( fd, FIONBIO, &nonblocking );
        #else
        retval = fcntl( fd, F_SETFL, fcntl( fd, F_GETFL ) | O_NONBLOCK );
        #endif
        assert( retval != SOCKET_ERROR );
        #ifdef WIN32
        retval = setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, (const char*)&reuseaddr, sizeof( reuseaddr ) );
        #else
        retval = setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, (const void*)&reuseaddr, sizeof( reuseaddr ) );
        #endif
        assert( retval != SOCKET_ERROR );
        addr.sin_family = AF_INET;
        addr.sin_port = htons( port );
        addr.sin_addr.s_addr = INADDR_ANY;
        retval = bind( fd, (sockaddr*)&addr, addrlen );
        assert( retval != SOCKET_ERROR );
        
        connectedClientAddrValid = false;
    }
    #ifdef WIN32
    addr2.sin_addr.s_addr = inet_addr( client );
    addrmask.sin_addr.s_addr = inet_addr( mask );
    #else
    retval = inet_pton( AF_INET, client, &addr2.sin_addr );
    assert( retval > 0 );
    retval = inet_pton( AF_INET, mask, &addrmask.sin_addr );
    assert( retval > 0 );
    #endif
    while( 1 )
    {
        char buf[10];
        retval = ::recvfrom( fd, buf, 10, MSG_PEEK, (sockaddr*)&addr, &addrlen );
        #ifdef WIN32
        int lerror = WSAGetLastError();
        assert( retval != SOCKET_ERROR || lerror == WSAEMSGSIZE || lerror == WSAENETDOWN || lerror == WSAEWOULDBLOCK );
        if( retval == SOCKET_ERROR && ( lerror == WSAENETDOWN || lerror == WSAEWOULDBLOCK ) )
        {
            if( wait )
            {
                Sleep( 100 );
                continue;
            }
            else
                break;
        }
        #else
        assert( retval != SOCKET_ERROR || errno == EAGAIN || errno == ENETDOWN || errno == EPROTO || errno == ENOPROTOOPT || errno == EHOSTDOWN || errno == ENONET || errno == EHOSTUNREACH || errno == EOPNOTSUPP || errno == ENETUNREACH );
        if( retval == SOCKET_ERROR && ( errno == EAGAIN || errno == ENETDOWN || errno == EPROTO || errno == ENOPROTOOPT || errno == EHOSTDOWN || errno == ENONET || errno == EHOSTUNREACH || errno == EOPNOTSUPP || errno == ENETUNREACH ) )
        {
            if( wait )
            {
                usleep( 100000 );
                continue;
            }
            else
                break;
        }
        #endif
        connectedClientAddr = addr;
        connectedClientAddrValid = true;
        if( !( ( addr.sin_addr.s_addr ^ addr2.sin_addr.s_addr ) & addrmask.sin_addr.s_addr ) ) // check address of client
        {
            this->lastReadTime = this->get_time( );
            fprintf( stderr, "...connected\n" );
            break;
        }
        else {
            fprintf( stderr, "rejecting connection from incorrect client\n" );
            this->close_client( );
        }
    }
    
    // repair options
    if( maskpos > 0 )
        options[ maskpos - 1 ] = '/';
    options[ i ] = ',';
    
    return ( this->connected( ) ? (int)fd : -1 );
}

// read from device (buf = buffer into which to read, count = max number of bytes to read,
//   maxInterval = time in seconds since last successful read after which to determine that
//   link has gone down, wait = whether to wait for reconnection if link has gone down,
//   returns number of bytes read or -1 on error)
//NOTE: SOCK_DGRAM does not return everything in buffer, only last message; this function compensates for that by looping the recv (but does not read part of a message from an allowed client, which would cause the rest of the message to be discarded)
ssize_t UdpServerCharDevice::read( void *buf, size_t count, double maxInterval /* = -1.0 */, bool wait /* = true */ )
{
    sockaddr_in recvAddr;
    socklen_t addrlen;
    ssize_t retval = 0;
    ssize_t sumretval = 0;
    bool peek = false;
    while( (size_t)sumretval < count )
    {
        if( !this->connected( ) )
        {
            //this->close_client( );
            if( this->open( wait ) == -1 && !wait )
            {
                retval = -1;
                break;
            }
            if( !this->connected( ) )
                continue;
        }
        addrlen = sizeof( recvAddr );
        #ifdef WIN32
        int lerror = 0;
        retval = ::recvfrom( fd, &(((char*)buf)[sumretval]), count - sumretval, peek ? MSG_PEEK : 0, (sockaddr*)&recvAddr, &addrlen ); //NOTE: was ::read()
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( ( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
            continue;
        }
        if( retval <= 0 && lerror == WSAEMSGSIZE )
            retval = count - sumretval;
        #else
        retval = ::recvfrom( fd, (void*)&(((char*)buf)[sumretval]), count - sumretval, peek ? MSG_PEEK : 0, (sockaddr*)&recvAddr, &addrlen ); //NOTE: was ::read()
        if( ( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) ) || ( retval <= 0 && maxInterval > 0.0 && ( this->get_time( ) - this->lastReadTime ) > maxInterval ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
            continue;
        }
        #endif
        if( retval <= 0 )
            break;
        assert( connectedClientAddrValid );
        if( recvAddr.sin_addr.s_addr ^ connectedClientAddr.sin_addr.s_addr )
        { // received data from different address than the one that first connected
            if( !peek )
                fprintf( stderr, "rejecting connection from additional client\n" );
            retval = 0;
        }
        if( peek )
        {
            if( (size_t)retval > ( count - (size_t)sumretval ) )
                break;
        }
        else
        { // data is OK
            sumretval += retval;
        }
        peek = !peek;
    }
    if( sumretval > 0 )
        retval = sumretval;
    else if( retval == SOCKET_ERROR )
        retval = -1;
    if( retval > 0 )
        this->lastReadTime = this->get_time( );
    return retval;
}

// write to device (buf = buffer from which to write, count = number number of bytes to write,
//   wait = whether to wait for reconnection if link has gone down, returns number of bytes
//   written or -1 on error)
ssize_t UdpServerCharDevice::write( const void *buf, size_t count, bool wait /*= true*/ )
{
    ssize_t retval;
    while( 1 )
    {
        if( !this->connected( ) )
        {
            //this->close_client( );
            if( this->open( wait ) == -1 && !wait )
            {
                retval = -1;
                break;
            }
            if( !this->connected( ) )
                continue;
        }
        #ifdef WIN32
        int lerror;
        retval = ::sendto( fd, (char*)buf, count, 0, (sockaddr*)&connectedClientAddr, sizeof( connectedClientAddr ) );
        if( retval == SOCKET_ERROR )
            lerror = WSAGetLastError();
        if( retval == SOCKET_ERROR && ( lerror == WSAENETRESET || lerror == WSAENOTCONN || lerror == WSAENOTSOCK || lerror == WSAEHOSTUNREACH || lerror == WSAECONNABORTED || lerror == WSAECONNRESET || lerror == WSAETIMEDOUT ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #else
        retval = ::sendto( fd, buf, count, MSG_NOSIGNAL, (sockaddr*)&connectedClientAddr, sizeof( connectedClientAddr ) );
        if( retval == SOCKET_ERROR && ( errno == EPIPE || errno == ECONNRESET || errno == EBADF || errno == ENOTCONN || errno == ENOTSOCK ) )
        {
            this->close_client( );
            if( this->open( wait ) == -1 && !wait )
                break;
        }
        else
            break;
        #endif
    }
    #ifdef WIN32
    if( retval == SOCKET_ERROR )
        retval = -1;
    #endif
    return retval;
}

// close device, called last time by destructor (returns 0 on success or -1 on error)
int UdpServerCharDevice::close( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    connectedClientAddrValid = false;
    #ifdef WIN32
    if( this->startedWSA )
    {
        WSACleanup( );
        this->startedWSA = false;
    }
    #endif
    return retval;
}

// close current client (returns 0 on success or -1 on error)
int UdpServerCharDevice::close_client( void )
{
    int retval = 0;
    if( fd != INVALID_SOCKET )
    {
        #ifdef WIN32
        retval = closesocket( fd );
        if( retval == SOCKET_ERROR )
            retval = -1;
        #else
        retval = ::close( fd );
        #endif
        fd = INVALID_SOCKET;
    }
    connectedClientAddrValid = false;
    return retval;
}

// whether currently connected
bool UdpServerCharDevice::connected( void )
{
    return ( fd != INVALID_SOCKET && connectedClientAddrValid );
}

// open device, called first time by constructor (wait = whether to wait for connection,
//   returns -1 on error)
int CharDeviceWrapper::open( bool wait )
{
    //assert( sizeof( HEADER8 ) == HEADER8::encoded_byte_count + HEADER8::pad_byte_count );
    //assert( sizeof( SerialChopperStateData ) == SerialChopperStateData::encoded_byte_count + SerialChopperStateData::pad_byte_count );
    
    assert( options );
    
    int i;
    int len = strlen( options );
    
    // find IO type
    for( i = 0; i < len && options[ i ] != ','; i++ );
    assert( i < len );
    options[ i ] = '\0';
    if( strcmp( options, "serial" ) == 0 )
    {
        options[ i ] = ',';
        charDevice = new SerialCharDevice( &options[ i + 1 ], wait );
    }
    else if( strcmp( options, "tcp" ) == 0 )
    {
        options[ i ] = ',';
        charDevice = new TcpCharDevice( &options[ i + 1 ], wait );
    }
    else if( strcmp( options, "tcplisten" ) == 0 || strcmp( options, "tcpserver" ) == 0 )
    {
        options[ i ] = ',';
        charDevice = new TcpServerCharDevice( &options[ i + 1 ], wait );
    }
    else if( strcmp( options, "udp" ) == 0 )
    {
        options[ i ] = ',';
        charDevice = new UdpCharDevice( &options[ i + 1 ], wait );
    }
    else if( strcmp( options, "udplisten" ) == 0 || strcmp( options, "udpserver" ) == 0 )
    {
        options[ i ] = ',';
        charDevice = new UdpServerCharDevice( &options[ i + 1 ], wait );
    }
    else
    {
        assert( strcmp( options, "serial" ) == 0 || strcmp( options, "tcp" ) == 0 || strcmp( options, "tcplisten" ) == 0 || strcmp( options, "tcpserver" ) == 0 || strcmp( options, "udp" ) == 0 || strcmp( options, "udplisten" ) == 0 || strcmp( options, "udpserver" ) == 0);
        //options[ i ] = ',';
    }
    
    return 0;
}

// read from device (buf = buffer into which to read, count = max number of bytes to read,
//   maxInterval = time in seconds since last successful read after which to determine that
//   link has gone down, wait = whether to wait for reconnection if link has gone down,
//   returns number of bytes read or -1 on error)
ssize_t CharDeviceWrapper::read( void *buf, size_t count, double maxInterval /* = -1.0 */, bool wait /* = true */ )
{
    return charDevice->read( buf, count, maxInterval, wait );
}

// write to device (buf = buffer from which to write, count = number number of bytes to write,
//   wait = whether to wait for reconnection if link has gone down, returns number of bytes
//   written or -1 on error)
ssize_t CharDeviceWrapper::write( const void *buf, size_t count, bool wait /*= true*/ )
{
    return charDevice->write( buf, count, wait );
}

// close device, called last time by destructor (returns 0 on success or -1 on error)
int CharDeviceWrapper::close( void )
{
    if( charDevice )
        delete charDevice;
    return 0;
}

// initialization (device = device options string, wait = whether to wait for connection)
void BufferedCharDevice::init( const char *device, bool wait /* = true */ )
{
    charDevice = new CharDeviceWrapper( device, wait );
    
    circbufSize = 256;
    circbuf = new BYTE[ circbufSize ];
    circbufStart = 0;
    circbufNext = 0;
    circbufLength = 0;
    
    //numUpdates = 0;
}

// cleanup
void BufferedCharDevice::cleanup( void )
{
    if( charDevice )
        delete charDevice;
            
    delete[] circbuf;
}

// read bytes into circular buffer (maxInterval = time in seconds since last successful
//    read after which to determine that link has gone down, wait = whether to wait for
//    reconnection if link has gone down)
void BufferedCharDevice::read_into_buffer( double maxInterval /* = -1.0 */, bool wait /* = true */ )
{
    int res, endspace;
    
    while( ( res = charDevice->read( /*ser_fd,*/ &circbuf[ circbufNext ], ( endspace = circbufSize - (int)circbufNext ), maxInterval, wait ) ) > 0 )
    {
        //fprintf( stderr, "read %d bytes\n", res );
        assert( res >= 0 && res <= circbufSize );
        assert( res <= endspace );
        assert( (int)circbufNext + res <= circbufSize );
        circbufNext += (BYTE)res;
        circbufLength += res;
        if( circbufLength > circbufSize )
        {
            circbufStart = circbufNext;
            circbufLength = circbufSize;
        }
        if( res < endspace )
            break;
    }
}


#ifdef CHARDEVICE_MAIN

// main function for standalone compilation (for testing)
int main( int argc, char* argv[] )
{
    CharDeviceWrapper *wrapper;
    char buf[ 1024 ];
    int nbytes;
    char *retval;
    
    if( argc < 2 )
    {
        fprintf( stderr, "USAGE: chardevice device\n" );
        return 0;
    }
    
    wrapper = new CharDeviceWrapper( argv[ 1 ], false );
    
    while( 1 )
    {
        nbytes = wrapper->read( buf, 1023/*, -1.0, false*/ );
        if( nbytes > 0 )
        {
            buf[ nbytes ] = '\0';
            puts( buf );
        }
        retval = fgets( buf, 1024, stdin );
        if( retval != 0 && ( nbytes = strlen( buf ) ) > 0 )
            wrapper->write( buf, nbytes, false );
    }
    
    delete wrapper;

    return 0;
}

#endif


#endif // __CHARDEVICE_C__
